from dataclasses import dataclass
from enum import Enum
from typing import Optional

from .option import Option


class Variant(Option, Enum):
    """Lichess chess variants"""

    STANDARD = Option("Standard", "standard")
    CRAZYHOUSE = Option("Crazyhouse", "crazyhouse")
    CHESS960 = Option("Chess 960", "chess960")
    KING_OF_THE_HILL = Option("King of the Hill", "kingOfTheHill")
    THREE_CHECK = Option("Three-check", "threeCheck")
    ANTICHESS = Option("Antichess", "antichess")
    ATOMIC = Option("Atomic", "atomic")
    HORDE = Option("Horde", "horde")
    RACING_KINGS = Option("Racing Kings", "racingKings")


class TimeMode(Option, Enum):
    """Lichess time modes / speeds"""

    ULTRABULLET = Option("Ultra-Bullet", "ultraBullet")
    BULLET = Option("Bullet", "bullet")
    BLITZ = Option("Blitz", "blitz")
    RAPID = Option("Rapid", "rapid")
    CLASSICAL = Option("Classical", "classical")
    CORRESPONDENCE = Option("Correspondence", "correspondence")


class Color(Option, Enum):
    """Lichess possible colors for the first player to join a match"""

    WHITE = Option("White", "white")
    BLACK = Option("Black", "black")
    RANDOM = Option("Random", "random")


class TimeControlType(Option, Enum):
    """Lichess time control types"""

    UNLIMITED = Option("Unlimited", "unlimited")
    CLOCK = Option("Clock", "clock")


@dataclass
class TimeControl:
    """
    Class representing a time control

    Attributes
    ----------

    type: :class:`TimeControlType`
        The type of time control (clock or unlimited)
    limit: Optional[:class:`int`]
        The clock time limit in seconds if real-time
    increment: Optional[:class:`int`]
        The clock increment in seconds if real-time
    show: Optional[:class:`str`]
        The time control string for display (eg. "5+2")
    """

    type: TimeControlType
    limit: Optional[int]
    increment: Optional[int]
    show: Optional[str]

    @classmethod
    def from_data(cls: "TimeControl", data: dict) -> "TimeControl":
        """Create a time control from a dictionary"""
        return cls(
            type=TimeControlType.find_by_data(data["type"]),
            limit=data.get("limit", None),
            increment=data.get("increment", None),
            show=data.get("show", None),
        )


@dataclass
class User:
    """
    Class to represent a Lichess user

    Attributes
    ----------

    id: str
        The user's Lichess ID
    name: str
        The user's Lichess username
    online: Optional[bool]
        Whether the user is online
    provisional: Optional[bool]
        Whether the user is provisional
    rating: Optional[:class:`int`]
        The user's Lichess rating
    title: Optional[:class:`str`]
        The user's Lichess title
    """

    id: str
    name: str
    online: Optional[bool]
    provisional: Optional[bool]
    rating: Optional[int]
    title: Optional[str]

    @classmethod
    def from_data(cls: "User", data: dict) -> "User":
        """Create a :class:`User` object from a dictionary of data

        Parameters
        ----------
        data: :class:`dict`
            A dictionary of data to create the :class:`User` object from

        Returns
        -------
        :class:`User`
            A :class:`User` object with the data from the dictionary
        """
        return cls(
            id=data["id"],
            name=data["username"],
            online=data.get("online", None),
            provisional=data.get("provisional", None),
            rating=data.get("rating", None),
            title=data.get("title", None),
        )



from typing import List, NamedTuple, TypeVar

T = TypeVar("T")


class Option(NamedTuple):
    description: str
    data: T

    def __str__(self):
        return self.description

    @classmethod
    def find_by_description(cls, description: str) -> "Option":
        """Get an option from a description

        Examples:
        ``Variant.find_by_description("Antichess")``
        ``Color.find_by_description("White")``
        ``TimeMode.find_by_description("Real-time")``
        """
        options: List[Option] = list(cls)
        for option in options:
            if description == option.description:
                return option
        raise ValueError("Unknown option")

    @classmethod
    def find_by_data(cls, data: str) -> "Option":
        """Get an option from its data

        Examples:
        ``Variant.find_by_data("antichess")``
        ``Color.find_by_data("white")``
        ``TimeMode.find_by_data("ultraBullet")``
        """
        options: List[Option] = list(cls)
        for option in options:
            if data == option.data:
                return option
        raise ValueError("Unknown option")

    @classmethod
    def find(cls, key: str) -> "Option":
        """Get an option from description or data

        Examples:
        ``Variant.find("Antichess")``
        ``Variant.find("ultraBullet")``
        ``TimeMode.find("Real-time")``
        ``TimeMode.find("realTime")``
        """
        try:
            return cls.find_by_description(key)
        except ValueError:
            return cls.find_by_data(key)



class BaseError(RuntimeError):
    """Base exception for exceptions caused by this package"""

    @property
    def message(self):
        raise NotImplementedError

    def __repr__(self):
        return self.message

    def __str__(self):
        return self.message


class HttpError(BaseError):
    """Exception caused by requests that were unsuccessful"""

    def __init__(
        self, status_code: int, reason: str, endpoint: str, response_text: str
    ):
        self.status_code = status_code
        self.reason = reason
        self.endpoint = endpoint
        self.response_text = response_text

    @property
    def message(self):
        return f"{self.status_code} {self.reason} {self.endpoint}\n{self.response_text}"


class BadArgumentError(BaseError):
    """Exception caused by invalid arguments"""

    def __init__(self, description: str):
        self.description = description

    @property
    def message(self):
        return self.description



from .match import Match, RealTimeMatch, CorrespondenceMatch, UnlimitedMatch

__version__ = "1.0.0"

__all__ = [
    "__version__",
    "Match",
    "RealTimeMatch",
    "CorrespondenceMatch",
    "UnlimitedMatch",
]



# /usr/bin/env python
import os
import re
from typing import List
from setuptools import setup
from setuptools.command.test import Command


class LintCommand(Command):
    """
    A copy of flake8's Flake8Command
    """

    description = "Run flake8 on modules registered in setuptools"
    user_options: List[str] = []

    def initialize_options(self):
        # must override
        pass

    def finalize_options(self):
        # must override
        pass

    def distribution_files(self):
        if self.distribution.packages:
            for package in self.distribution.packages:
                yield package.replace(".", os.path.sep)

        if self.distribution.py_modules:
            for filename in self.distribution.py_modules:
                yield "%s.py" % filename

    def run(self):
        from flake8.api.legacy import get_style_guide

        flake8_style = get_style_guide(config_file="setup.cfg")
        paths = self.distribution_files()
        report = flake8_style.check_files(paths)
        raise SystemExit(report.total_errors > 0)


version = ""
with open(os.path.join("play_lichess", "__init__.py")) as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

if not version:
    raise RuntimeError("version is not set")

long_description = ""
if os.path.exists("README.md"):
    with open("README.md", "r") as f:
        long_description = f.read()

requirements = []
if os.path.exists("requirements.txt"):
    with open("requirements.txt", "r") as f:
        requirements = f.read().splitlines()

setup(
    name="play-lichess",
    version=version,
    author="Jonah Lawrence",
    author_email="jonah@freshidea.com",
    description="Module for creating match links on Lichess that players can join",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DenverCoder1/play-lichess",
    project_urls={
        "Bug Tracker": "https://github.com/DenverCoder1/play-lichess/issues",
    },
    packages=["play_lichess"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
    ],
    python_requires=">=3.8",
    install_requires=[requirements],
    cmdclass={
        "lint": LintCommand,
    },
)



import json
from dataclasses import dataclass
from typing import Any, Coroutine, Optional

import aiohttp

from .exceptions import BadArgumentError, HttpError
from .types import Color, TimeControl, TimeMode, User, Variant


@dataclass
class Match:
    """Class for storing results about a created match link

    Attributes
    ----------
    challenge_id: :class:`str`
        The challenge id of the match
    challenge_url: :class:`str`
        The url of the match
    status: :class:`str`
        The status of the match
    challenger: Optional[:class:`str`]
        The challenger of the match
    dest_user: Optional[:class:`str`]
        The destination user of the match
    variant: :class:`Variant`
        The variant of the match
    rated: :class:`bool`
        Whether the match is rated
    speed: :class:`TimeMode`
        The speed of the match
    time_control: :class:`TimeControl`
        The time control of the match
    color: :class:`Color`
        The color of the player who starts the match
    url_white: :class:`str`
        The url of the white player
    url_black: :class:`str`
        The url of the black player
    """

    challenge_id: str
    challenge_url: str
    status: str
    challenger: Optional[User] = None
    dest_user: Optional[User] = None
    variant: Variant = Variant.STANDARD
    rated: bool = False
    speed: TimeMode = TimeMode.CORRESPONDENCE
    time_control: Optional[TimeControl] = None
    color: Color = Color.RANDOM
    url_white: Optional[str] = None
    url_black: Optional[str] = None
    name: Optional[str] = None
    _data: Optional[dict] = None

    @classmethod
    def from_data(cls: "Match", data: dict, name: Optional[str] = None) -> "Match":
        """Create a :class:`Match` object from a dictionary of data

        Parameters
        ----------
        data: :class:`dict`
            A dictionary of data to create the :class:`Match` object from
        name: Optional[:class:`str`]
            The name of the match

        Returns
        -------
        :class:`Match`
            A :class:`Match` object with the data from the dictionary
        """
        return cls(
            challenge_id=data["challenge"]["id"],
            challenge_url=data["challenge"]["url"],
            status=data["challenge"]["status"],
            challenger=(
                User.from_data(data["challenge"]["challenger"])
                if data["challenge"]["challenger"]
                else None
            ),
            dest_user=(
                User.from_data(data["challenge"]["destUser"])
                if data["challenge"]["destUser"]
                else None
            ),
            variant=Variant.find_by_data(data["challenge"]["variant"]["key"]),
            rated=data["challenge"]["rated"],
            speed=TimeMode.find_by_data(data["challenge"]["speed"]),
            time_control=TimeControl.from_data(data["challenge"]["timeControl"]),
            color=Color.find_by_data(data["challenge"]["color"]),
            url_white=data["urlWhite"],
            url_black=data["urlBlack"],
            name=name,
            _data=data,
        )

    @classmethod
    async def create(
        cls: "Match",
        *,
        rated: bool = False,
        clock_limit: Optional[int] = 300,
        clock_increment: Optional[int] = 0,
        days: Optional[int] = None,
        variant: Variant = Variant.STANDARD,
        fen: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Coroutine[Any, Any, "Match"]:
        """Start a match that two players can join

        :param rated: :class:`bool`
            Game is rated and impacts players ratings
        :param clock_limit: :class:`int`
            Clock initial time in seconds. Leave blank for a correspondence or unlimited match.
            If specified, must be between 0 and 10800 seconds.
        :param clock_increment: :class:`int`
            Clock increment in seconds. Leave blank for a correspondence or unlimited match.
            If specified, must be between 0 and 180 seconds.
        :param days: :class:`int`
            Days per turn for correspondence matches. Leave blank for a live or unlimited match.
            If specified, must be between 1, 2, 3, 5, 7, 10, or 14 days.
        :param variant: :class:`Variant`
            The variant of the match (STANDARD, ANTICHESS, CHESS960, etc.)
            The default is STANDARD
        :param fen: :class:`str`
            Custom initial position (in FEN). Variant must be standard, and the game cannot be rated.
            The default position is "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"
        :param name: :class:`str`
            Optional name for the challenge that players will see on the challenge page.

        Returns
        -------
        :class:`Match`
            A :class:`Match` object with the data from the API

        Raises
        ------
        :class:`BadArgumentError`
            If days is set and clock_limit or clock_increment is also set.
            If one of clock_limit or clock_increment is set but the other is not.
        :class:`HttpError`
            If the HTTP request fails, for example:
            If clock_limit or clock_increment is not in the valid range.
            If clock_limit and clock_increment are both set to 0.
            If days is not in the valid range.
            If rated is set, but there is no time control.
            If a rate-limit or server error occurs.
        """
        if days and (clock_limit or clock_increment):
            raise BadArgumentError(
                "days cannot be set with clock_limit or clock_increment"
            )
        if not days and (clock_limit is None) ^ (clock_increment is None):
            raise BadArgumentError(
                "Both clock_limit and clock_increment must be specified or neither"
            )
        if fen is not None:
            if variant != Variant.STANDARD:
                raise BadArgumentError(
                    "fen can only be specified for STANDARD variants"
                )
            if rated:
                raise BadArgumentError("fen can only be specified for unrated games")

        endpoint_url = "https://lichess.org/api/challenge/open"
        params = {
            "rated": rated,
            "clock.limit": clock_limit,
            "clock.increment": clock_increment,
            "variant": variant.value.data,
            "days": days,
            "fen": fen,
            "name": name,
        }
        data = json.dumps({k: v for k, v in params.items() if v is not None})
        headers = {"User-Agent": "play-lichess", "Content-Type": "application/json"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                endpoint_url, data=data, headers=headers
            ) as response:
                if response.status != 200:
                    raise HttpError(
                        status_code=response.status,
                        reason=response.reason,
                        endpoint=endpoint_url,
                        response_text=await response.json(),
                    )
                return cls.from_data(await response.json(), name)


class RealTimeMatch(Match):
    @classmethod
    async def create(
        cls: "RealTimeMatch",
        *,
        rated: bool = False,
        clock_limit: Optional[int] = 300,
        clock_increment: Optional[int] = 0,
        variant: Variant = Variant.STANDARD,
        fen: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Coroutine[Any, Any, "Match"]:
        return await super().create(
            rated=rated,
            clock_limit=clock_limit,
            clock_increment=clock_increment,
            variant=variant,
            fen=fen,
            name=name,
        )


class CorrespondenceMatch(Match):
    @classmethod
    async def create(
        cls: "CorrespondenceMatch",
        *,
        rated: bool = False,
        days: Optional[int] = 1,
        variant: Variant = Variant.STANDARD,
        fen: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Coroutine[Any, Any, "Match"]:
        return await super().create(
            rated=rated,
            clock_limit=None,
            clock_increment=None,
            days=days,
            variant=variant,
            fen=fen,
            name=name,
        )


class UnlimitedMatch(Match):
    @classmethod
    async def create(
        cls: "UnlimitedMatch",
        *,
        variant: Variant = Variant.STANDARD,
        fen: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Coroutine[Any, Any, "Match"]:
        return await super().create(
            rated=False,
            clock_limit=None,
            clock_increment=None,
            days=None,
            variant=variant,
            fen=fen,
            name=name,
        )
